//********************************************************************/
//* Copyright (C) 2005-2011                                          */
//* MEL O'CAT  X178G243 (at) yahoo (dot) com                         */
//* License terms: GNU General Public License Version 2              */
//*                or any later version                              */
//********************************************************************/
//*4567890123456 (71-character line to adjust editor window) 23456789*/

/*
 *  BatchFramework.java  0.09 11/01/2011
 *
 *  Sep-25-2005:
 *      -->Start counting RunParmFile lines at 1. Duh.
 *  Dec-03-2005:
 *      -->Add ProofAsst stuff
 *      -->Added exception/error message and stack trace
 *         print directly to System.err in runIt() because
 *         NullPointException was just dumping with the
 *         helpful message "null".
 *  Sep-03-2006:
 *      -->Add TMFF stuff
 *  Jun-01-2007 - Version 0.06
 *      -->OutputVerbosity RunParm stuff.
 *  Sep-01-2007 - Version 0.07
 *      -->Add WorkVarBoss.
 *  Aug-01-2008 - Version 0.08
 *      -->Add SvcBoss.
 *      -->Make sure LogicalSystemBoss is at the end of the
 *         Boss list. This gives other bosses a chance to
 *         see the LoadFile command and re-initialize
 *         themselves (rare situation: only if multiple
 *         sets of commands in a single RunParm file.)
 *      -->Add TheoremLoaderBoss
 *  Nov-01-2011 - Version 0.09
 *      -->Add GMFFBoss.
 *      -->Add GMFFException.
 *      -->Modified for mmj2 Paths Enhancement
 *      -->Added code for MMJ2FailPopupWindow
 */

package mmj.util;

import java.io.*;
import java.util.*;
import mmj.mmio.*;
import mmj.lang.*;
import mmj.verify.*;
import mmj.pa.*;
import mmj.tmff.*;
import mmj.svc.*;
import mmj.tl.*;
import mmj.gmff.*;

/**
 *  BatchFramework is a quick hack to run mmj2 without
 *  the JUnit training wheels.
 *  <p>
 *  An example of using this "framework" is BatchMMJ2
 *  which sub-classes BatchFramework and is invoked
 *  via "main(String[] args)", passing those parameters
 *  in turn, to BatchFramework.
 *  <p>
 *  A RunParmFile is used to provide flexibility for
 *  many, many parameters in the future, and to allow
 *  for files using different code sets.
 *  <p>
 *  This code is experimental and goofy looking :)
 *  No warranty provided... The theme is that the
 *  order of input RunParmFile commands is unknown
 *  and therefore, each function must not make any
 *  assumptions about what has been done previously.
 *  <p>
 *  "Boss" classes are used to manage "state"
 *  information, and "get" commands build
 *  objects as needed, invoking functions under the
 *  control of other Bosses to obtain needed objects
 *  (which are *not* to be retained given the dynamic
 *  flux of state information.) This all seems very
 *  inefficient but the overhead is very small
 *  compared to the amount of work performed by each
 *  RunParmFile command.
 *  <p>
 *  A list of Bosses in use during a given run is
 *  maintained and each input RunParm line is
 *  sent to each Boss, in turn, which may or may
 *  not do anything with the command. Upon exit,
 *  a Boss returns "consumed" to indicate that
 *  the RunParm need not be sent to any other
 *  Bosses, that the job is done. Adding a new
 *  function, such as "Soundness checking" should
 *  be simple.
 *  <p>
 *  An alternate way to use BatchFramework is to
 *  instantiate one but instead of executing "runIt",
 *  invoke initializeBatchFramework() and then
 *  directly call routines in the various Boss
 *  classes, passing them hand-coded RunParmArrayEntry
 *  objects built using the UtilConstants RunParm
 *  name and value literals.. In other words, don't
 *  use a RunParmFile. This approach provides a
 *  short-cut for invoking mmj2 functions, to which
 *  new functions could be easily applied.
 *
 */
public abstract class BatchFramework {

    protected    boolean batchFrameworkInitialized
                                   = false;

	protected    CommandLineArguments
	                               commandLineArguments;

    /*friendly*/ Paths             paths;

    /*friendly*/ MMJ2FailPopupWindow
                                   mmj2FailPopupWindow;

    /*friendly*/ boolean           displayMMJ2FailPopupWindow
                                =
		UtilConstants.DISPLAY_MMJ2_FAIL_POPUP_WINDOW_DEFAULT;

    protected    RunParmFile       runParmFile;
    /*friendly*/ int               runParmCnt;

	/*friendly*/ String            currentRunParmCommand;

    /*friendly*/ String            runParmExecutableCaption;
    /*friendly*/ String            runParmCommentCaption;

    /*friendly*/ ArrayList         bossList;

    /*friendly*/ OutputBoss        outputBoss;

    /*friendly*/ LogicalSystemBoss logicalSystemBoss;

    /*friendly*/ VerifyProofBoss   verifyProofBoss;

    /*friendly*/ GrammarBoss       grammarBoss;

    /*friendly*/ ProofAsstBoss     proofAsstBoss;

    /*friendly*/ TMFFBoss          tmffBoss;

    /*friendly*/ WorkVarBoss       workVarBoss;

    /*friendly*/ SvcBoss           svcBoss;

    /*friendly*/ TheoremLoaderBoss theoremLoaderBoss;

    /*friendly*/ GMFFBoss          gmffBoss;


    /**
     *  Default Constructor.
     */
    public BatchFramework() {
    }

    /**
     *  Initialize BatchFramework with Boss list and
     *  any captions that may have been overridden.
     *
     *  The purpose of doing this here is to allow a
     *  BatchFramework to be constructed without
     *  executing any runparms from a file: every
     *  "doRunParm" function is public and can be
     *  called from a program (assuming the program
     *  can create a valid RunParmArrayEntry to
     *  provide RunParm option values.) This provides
     *  a shortcut to invoking complicated mmj2 functions
     *  that would otherwise require lots of setup
     *  and parameters.
     */
    public void initializeBatchFramework() {
        batchFrameworkInitialized = true;
        bossList                  = new ArrayList();

        outputBoss                = new OutputBoss(this);
        addBossToBossList(outputBoss);

        verifyProofBoss           = new VerifyProofBoss(this);
        addBossToBossList(verifyProofBoss);

        grammarBoss               = new GrammarBoss(this);
        addBossToBossList(grammarBoss);

        proofAsstBoss             = new ProofAsstBoss(this);
        addBossToBossList(proofAsstBoss);

        tmffBoss                  = new TMFFBoss(this);
        addBossToBossList(tmffBoss);

        workVarBoss               = new WorkVarBoss(this);
        addBossToBossList(workVarBoss);

        svcBoss                   = new SvcBoss(this);
        addBossToBossList(svcBoss);

        theoremLoaderBoss        = new TheoremLoaderBoss(this);
        addBossToBossList(theoremLoaderBoss);

        gmffBoss                 = new GMFFBoss(this);
        addBossToBossList(gmffBoss);


        addAnyExtraBossesToBossList();

        //NOTE: LogicalSystemBoss should be at the end
        //      because the "LoadFile" RunParm is a signal
        //      to many other the other bosses that they
        //      need to re-initialize their state (for
        //      example, Grammar requires reinitialization
        //      for new input .mm files.)
        logicalSystemBoss         = new LogicalSystemBoss(this);
        addBossToBossList(logicalSystemBoss);

        setRunParmExecutableCaption();
        setRunParmCommentCaption();

        mmj2FailPopupWindow     =
        	new MMJ2FailPopupWindow(this,
        	                        displayMMJ2FailPopupWindow);
    }

    /**
     *  Uses command line run parms to build <code>RunParmFile</code>
     *  and <code>Paths</code> objects, performs other initialization
     *  and processes each RunParmFile line.
     *  <p>
     *  The <code>MMJ2FailPopupWindow</code> object is initialized
     *  in startupMode and gathers/displays error messages during
     *  startup logic. See <code>OutputBoss.printAndClearMessages()</code>
     *  which does the gathering and displaying of startup errors.
     *  Abnormal termination -- "Fail" -- error messages are
     *  displayed here by calling the MMJ2 Fail Popup Window directly.
     *
     *  @param args command line parms for RunParmFile constructor.
     *              (See CommandLineArguments.java for detailed doc
     *
     *  @return return code 0 if BatchFramework was successful
     *          (however many mmj/Metamath errors were found),
     *          or 16, if BatchFramework failed to complete
     *          (probably due to a RunParmFile error.)
     */
    public int runIt(String[] args) {

        int    retCd            = 0;
        String failMessage      = null;

        if (!batchFrameworkInitialized) {
            initializeBatchFramework();
        }

        try {
			commandLineArguments
								=
				new CommandLineArguments(args);

			paths               =
				commandLineArguments.getPaths();

			runParmFile         =
				commandLineArguments.getRunParmFile();

			displayMMJ2FailPopupWindow
			                    =
				commandLineArguments.getDisplayMMJ2FailPopupWindow();

			mmj2FailPopupWindow.
				setEnabled(
					displayMMJ2FailPopupWindow);

			mmj2FailPopupWindow.initiateStartupMode();

		}
        catch(Exception e) {
			failMessage         =
				UtilConstants.ERRMSG_RUNPARM_FILE_BOGUS_1
                + e.getMessage();
            System.err.println(
				failMessage);
            retCd               = 16;
        }

		if (retCd == 0) {
			try {
				while (runParmFile.hasNext()) {
					executeRunParmCommand(runParmFile.next());
				}
			}
			catch (Exception e) {
				failMessage         = e.getMessage();
				System.err.println(failMessage);
				e.printStackTrace(System.err);
				retCd               = 16;
				try {
					outputBoss.sysErrPrintln(failMessage);
				}
				catch(IOException f) {
					failMessage     = failMessage
									+ f.getMessage();
					System.err.println(failMessage);
				}
			}
			catch (Error e) {
				failMessage         = e.getMessage();
				retCd               = 16;
				System.err.println(failMessage);
				e.printStackTrace(System.err);
			}
		}

		if (failMessage != null) {
			mmj2FailPopupWindow.displayFailMessage(failMessage);
		}

		outputBoss.close();

        return retCd;
    }

    /**
     *  Add a processing Boss to the Boss List.
     *
     *  @param boss
     */
    public void addBossToBossList(Boss boss) {
        bossList.add(boss);
    }

    /**
     *  Placeholder to be overridden to dynamically add
     *  a new processing Boss to the list.
     *  <p>
     *  Override this to include extra processing. Each
     *  Boss on the bossList will get a chance to process
     *  each RunParmFile line that hasn't already been
     *  "consumed" by other bosses.
     */
    public void addAnyExtraBossesToBossList() {
    }

    /**
     *  Override this to alter what prints out before
     *  an executable RunParmFile line is processed.
     */
    public void setRunParmExecutableCaption() {
        runParmExecutableCaption   =
            UtilConstants.ERRMSG_RUNPARM_EXECUTABLE_CAPTION;
    }

    /**
     *  Override this to alter what prints out for
     *  a RunParmFile comment line.
     */
    public void setRunParmCommentCaption() {
        runParmCommentCaption   =
            UtilConstants.ERRMSG_RUNPARM_COMMENT_CAPTION;
    }

    /**
     *  Processes a single RunParmFile line.
     *
     *  @param runParm RunParmFileLine parsed into a
     *         RunParmArrayEntry object.
     */
    public void executeRunParmCommand(RunParmArrayEntry runParm)
                            throws IllegalArgumentException,
                                   MMIOException,
                                   FileNotFoundException,
                                   IOException,
                                   VerifyException,
                                   TheoremLoaderException,
                                   GMFFException {


        boolean  consumed;
        Iterator iterator;
        ++runParmCnt;

		//capture this for use by MMJ2FailPopupWindow
        currentRunParmCommand   = runParm.name;

        if (runParm.commentLine != null) {
            printCommentRunParmLine(
                             runParmCommentCaption,
                             runParmCnt,
                             runParm);
        }
        else {
            printExecutableRunParmLine(
                             runParmExecutableCaption,
                             runParmCnt,
                             runParm);

            if (runParm.name.compareToIgnoreCase(
                UtilConstants.RUNPARM_JAVA_GARBAGE_COLLECTION)
                == 0) {
                System.gc();
            }
            else {
                consumed = false;
                iterator = bossList.iterator();
                while (iterator.hasNext()
                       &&
                       !consumed) {
                    consumed      =
                        ((Boss)iterator.next()
                            ).doRunParmCommand(runParm);
                }
                if (!consumed &&
                    runParm.name.compareToIgnoreCase(
                        UtilConstants.RUNPARM_CLEAR)
                    != 0) {
                    throw new IllegalArgumentException(
                        UtilConstants.ERRMSG_RUNPARM_NAME_INVALID_1
                        + runParm.name
                        + UtilConstants.ERRMSG_RUNPARM_NAME_INVALID_2);
                }
            }
        }
    }

    /**
     *  Compares the current RunParmArrayEntry name to an input
     *  command.
     *  <p>
     *  @param command RunParm command to compare to the current
     *                 command name.
     *  @return true if equal, ignoring case, otherwise false.
     */
	public boolean isCurrentRunParmCommand(String command) {
		if (currentRunParmCommand != null
				&&
		    currentRunParmCommand.compareToIgnoreCase(command) == 0) {
			return true;
		}
		else {
			return false;
		}
	}

    /**
     *  Override this to change or eliminate the printout
     *  of each executable RunParmFile line.
     *
     *  @param caption to print
     *  @param cnt     RunParmFile line number
     *  @param runParm RunParmFile line parsed into
     *                 object RunParmArrayEntry.
     */
    public void printExecutableRunParmLine(
                                 String            caption,
                                 int               cnt,
                                 RunParmArrayEntry runParm)
                            throws IOException {
        outputBoss.sysOutPrintln(
                        caption
                        + cnt
                        + UtilConstants.ERRMSG_EQUALS_LITERAL
                        + runParm,
                        UtilConstants.RUNPARM_LINE_DUMP_VERBOSITY);
    }

    /**
     *  Override this to change or eliminate the printout
     *  of each Comment RunParmFile line.
     *
     *  @param caption to print
     *  @param cnt     RunParmFile line number
     *  @param runParm RunParmFile line parsed into
     *                 object RunParmArrayEntry.
     */
    public void printCommentRunParmLine(
                                 String            caption,
                                 int               cnt,
                                 RunParmArrayEntry runParm)
                            throws IOException {
        outputBoss.sysOutPrintln(
                        caption
                        + cnt
                        + UtilConstants.ERRMSG_EQUALS_LITERAL
                        + runParm,
                        UtilConstants.RUNPARM_LINE_DUMP_VERBOSITY);
    }

    /**
     *  Returns the canonical path name of the RunParmFile.
     *  <p>
     *  @return canonical path of RunParmFile or empty string
     *          if RunParmFile is null.
     */
	public String getRunParmFileAbsolutePath() {
		String s                = new String("");
		if (runParmFile != null) {
			s                   = runParmFile.getAbsolutePath();
		}
		return s;
	}
}
